using System.Collections.Generic;
using ImageProcessingGraph.Editor;
using UnityEditor.Experimental.GraphView;
using UnityEngine;
using UnityEngine.UIElements;

namespace AssetGraph.Core
{
    public class AssetGraphEdgeConnector : IEdgeConnectorListener
    {
        private readonly GraphViewChange graphViewChange;
        private readonly List<Edge> edgesToCreate;
        private readonly List<GraphElement> edgesToDelete;
        private readonly AssetGraphViewWindow window;

        public AssetGraphEdgeConnector(AssetGraphViewWindow window)
        {
            this.window = window;
            edgesToCreate = new List<Edge>();
            edgesToDelete = new List<GraphElement>();
            graphViewChange = new GraphViewChange
            {
                edgesToCreate = edgesToCreate
            };
        }

        public void OnDropOutsidePort(Edge edge, Vector2 position)
        {
            HandleSearchProvider(edge);

            RemoveExistingConnections(edge);

            if (edge.input?.node != null)
            {
                ((AssetGraphNodeEditor)edge.input.node)
                    .ToggleExposedVariable((AssetGraphPort)edge.input, true);
            }
        }

        public void OnDrop(GraphView graphView, Edge edge)
        {
            RemoveExistingConnections(edge);

            if (edge.input?.node != null)
            {
                ((AssetGraphNodeEditor)edge.input.node)
                    .ToggleExposedVariable((AssetGraphPort)edge.input, true);
            }

            UpdateConnections(graphView, edge);
        }

        private void HandleSearchProvider(Edge edge)
        {
            // Set the target for the search provider
            window.searchProvider.target = (VisualElement)window.focusController.focusedElement;

            var mousePos = window.ChangeCoordinatesTo(window, window.cachedMousePos - window.Window.position.position);
            var graphMousePosition = window.contentViewContainer.WorldToLocal(mousePos);

            var searchProviderInstance =
                ScriptableObject.CreateInstance<AssetGraphSearchProvider.CustomSearchProviderForEdge>();
            searchProviderInstance.Init(window.searchProvider, edge, (AssetGraphPort)edge.input,
                (AssetGraphPort)edge.output);

            SearchWindow.Open(
                new SearchWindowContext(GUIUtility.GUIToScreenPoint(Event.current.mousePosition)),
                searchProviderInstance
            );
        }

        private void RemoveExistingConnections(Edge edge)
        {
            // Collect and remove connections associated with the edge
            var connectionsToRemove = new List<GraphConnection>();
            foreach (var connection in window.asset.Connections)
            {
                if (connection.internalEdge == edge)
                {
                    connectionsToRemove.Add(connection);
                }
            }

            foreach (var connection in connectionsToRemove)
            {
                window.asset.Connections.Remove(connection);
            }
        }

        private void UpdateConnections(GraphView graphView, Edge edge)
        {
            // Clear the list of created and deleted edges
            edgesToCreate.Clear();
            edgesToDelete.Clear();

            // Add the current edge to edgesToCreate
            edgesToCreate.Add(edge);

            // Handle single-capacity input ports
            if (edge.input.capacity == Port.Capacity.Single)
            {
                foreach (var connection in edge.input.connections)
                {
                    if (connection != edge)
                    {
                        edgesToDelete.Add(connection);
                    }
                }
            }

            // Handle single-capacity output ports
            if (edge.output.capacity == Port.Capacity.Single)
            {
                foreach (var connection in edge.output.connections)
                {
                    if (connection != edge)
                    {
                        edgesToDelete.Add(connection);
                    }
                }
            }

            // Delete unwanted connections
            if (edgesToDelete.Count > 0)
            {
                graphView.DeleteElements(edgesToDelete);
            }

            // Create new connections
            var createdEdges = graphView.graphViewChanged != null
                ? graphView.graphViewChanged(graphViewChange).edgesToCreate
                : edgesToCreate;

            foreach (var createdEdge in createdEdges)
            {
                graphView.AddElement(createdEdge);
                edge.input.Connect(createdEdge);
                edge.output.Connect(createdEdge);
            }
        }
    }
}