﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Reflection;
using AssetGraph.Core.Attributes;
using AssetGraph.Core.GraphElements;
using AssetGraph.Nodes;
using UnityEditor;
using UnityEditor.Experimental.GraphView;
using UnityEditor.UIElements;
using UnityEngine;
using UnityEngine.UIElements;
using Input = AssetGraph.Core.Attributes.Input;

namespace AssetGraph.Core
{
    public class AssetGraphNodeEditor : Node
    {
        private readonly StyleSheet defaultStyleSheet;
        private readonly StyleSheet errorStyleSheet;
        private readonly NodeInfoAttribute info;

        public AssetGraphNode GraphNode { get; }
        public List<Port> AllPorts { get; } = new List<Port>();
        public List<Port> InputPorts { get; } = new List<Port>();
        public List<Port> OutputPorts { get; } = new List<Port>();

        private readonly AssetGraphViewWindow window;

        public AssetGraphNodeEditor(AssetGraphNode node, AssetGraphViewWindow window)
        {
            this.window = window;
            GraphNode = node;

            // Initialize node UI
            info = InitializeNode(node);

            // Collect input and output ports
            CreatePorts(node.GetType());

            // Load stylesheets
            defaultStyleSheet = LoadStyleSheet("Node.uss", "Packages/com.chromium.imageprocessingrah/Node.uss");
            errorStyleSheet = LoadStyleSheet("NodeError.uss", "Packages/com.chromium.imageprocessingrah/NodeError.uss");
            styleSheets.Add(defaultStyleSheet);

            // Set up error handling and cleanup
            SetUpErrorHandling();
        }

        /// <summary>
        /// Initializes the node, including title, name, and class lists.
        /// </summary>
        private NodeInfoAttribute InitializeNode(AssetGraphNode node)
        {
            var typeInfo = node.GetType();
            var nodeInfo = typeInfo.GetCustomAttribute<NodeInfoAttribute>();

            title = nodeInfo.Title;
            name = typeInfo.Name;

            foreach (var depth in nodeInfo.MenuItem.Split('/'))
            {
                AddToClassList(depth.ToLower().Replace(' ', '-'));
            }

            return nodeInfo;
        }

        /// <summary>
        /// Creates input and output ports for the node from its fields.
        /// </summary>
        private void CreatePorts(Type typeInfo)
        {
            var fields = typeInfo.GetFields(BindingFlags.Instance | BindingFlags.NonPublic | BindingFlags.Public);

            // Collect input and output fields
            var inputFields = fields.Where(f => f.GetCustomAttribute<Input>() != null).ToList();
            var outputFields = fields.Where(f => f.GetCustomAttribute<Output>() != null).ToList();

            // Create ports
            CreateInputPorts(inputFields);
            CreateOutputPorts(outputFields);
        }

        private void CreateInputPorts(List<FieldInfo> fields)
        {
            foreach (var field in fields)
            {
                var port = CreatePort(field, true, field.GetCustomAttribute<Input>().Label);
                InputPorts.Add(port);
                AllPorts.Add(port);

                inputContainer.Add(port);
                ExposeVariableToPort(port, field);
                ((AssetGraphPort)port).fieldInfo = field;
            }
        }

        private void CreateOutputPorts(List<FieldInfo> fields)
        {
            foreach (var field in fields)
            {
                var port = CreatePort(field, false, field.GetCustomAttribute<Output>().Label);
                OutputPorts.Add(port);
                AllPorts.Add(port);

                outputContainer.Add(port);
                ((AssetGraphPort)port).fieldInfo = field;
            }
        }

        private Port CreatePort(FieldInfo field, bool isInput, string label)
        {
            var port = AssetGraphPort.Create(window.edgeConnectorListener, isInput, field.FieldType);
            if (!string.IsNullOrEmpty(label)) port.portName = label;
            return port;
        }

        private void SetUpErrorHandling()
        {
            GraphNode.onFailed += () => styleSheets.Add(errorStyleSheet);

            window.asset.OnRun += () =>
            {
                if (styleSheets.Contains(errorStyleSheet))
                    styleSheets.Remove(errorStyleSheet);
            };
        }

        private StyleSheet LoadStyleSheet(string path, string fallbackPath)
        {
            var styleSheet = AssetDatabase.LoadAssetAtPath<StyleSheet>($"Assets/Unity Image Processing/{path}");
            return styleSheet ?? EditorGUIUtility.Load(fallbackPath) as StyleSheet;
        }

        /// <summary>
        /// Exposes a variable on the port for editing when it's not connected.
        /// </summary>
        public void ExposeVariableToPort(Port port, FieldInfo field)
        {
            var variableContainer = ((AssetGraphPort)port).ExposedPropertyContainer;

            if (variableContainer == null)
            {
                var newElement = CreatePropertyFieldForType(field.FieldType, field.GetValue(GraphNode));
                if (newElement != null)
                {
                    variableContainer = newElement;
                    ((AssetGraphPort)port).ExposedPropertyContainer = newElement;

                    // Add the edited element to the port
                    port.Add(newElement);
                    RegisterFieldChangeCallback(field, newElement);
                    variableContainer.style.display = DisplayStyle.Flex;
                }
            }

        }

        private void RegisterFieldChangeCallback(FieldInfo field, VisualElement element)
        {
            // Register value change callbacks for various field types
            switch (element)
            {
                case IntegerField intField:
                    if (field.FieldType == typeof(GreyscaleValue))
                    {
                        // Handle GreyscaleValue specifically
                        intField.RegisterValueChangedCallback(evt =>
                        {
                            var value = (GreyscaleValue)field.GetValue(GraphNode);

                            if (intField is GreyscaleField greyscaleField && value != null)
                            {
                                if (evt.newValue > greyscaleField.minMax.Item2)
                                    value.value = greyscaleField.minMax.Item2;
                                else if (evt.newValue < greyscaleField.minMax.Item1)
                                    value.value = greyscaleField.minMax.Item1;
                                else
                                    value.value = evt.newValue;

                                field.SetValue(GraphNode, value);
                                intField.SetValueWithoutNotify((int)value.value);

                            }
                        });
                    }
                    else
                    {
                        // Default case for IntegerField
                        intField.RegisterValueChangedCallback(evt => field.SetValue(GraphNode, evt.newValue));
                    }

                    break;

                case FloatField floatField:
                    floatField.RegisterValueChangedCallback(evt => field.SetValue(GraphNode, evt.newValue));
                    break;

                case Toggle boolField:
                    boolField.RegisterValueChangedCallback(evt => field.SetValue(GraphNode, evt.newValue));
                    break;

                case TextField stringField:
                    stringField.RegisterValueChangedCallback(evt => field.SetValue(GraphNode, evt.newValue));
                    break;

                case ColorField colorField:
                    colorField.RegisterValueChangedCallback(evt => field.SetValue(GraphNode, evt.newValue));
                    break;

                case Vector3Field vector3Field:
                    vector3Field.RegisterValueChangedCallback(evt => field.SetValue(GraphNode, evt.newValue));
                    break;

                case Vector2Field vector2Field:
                    vector2Field.RegisterValueChangedCallback(evt => field.SetValue(GraphNode, evt.newValue));
                    break;

                case EnumField enumField:
                    enumField.RegisterValueChangedCallback(evt => field.SetValue(GraphNode, evt.newValue));
                    break;

                case ObjectField objectField:
                    objectField.RegisterValueChangedCallback(evt => field.SetValue(GraphNode, evt.newValue));
                    break;
            }
        }

        private VisualElement CreatePropertyFieldForType(Type type, object value)
        {
            return type switch
            {
                { } when type == typeof(int) => new IntegerField { value = (int)value },
                { } when type == typeof(float) => new FloatField { value = (float)value },
                { } when type == typeof(bool) => new Toggle { value = (bool)value },
                { } when type == typeof(string) => new TextField { value = (string)value },
                { } when type == typeof(Color) => new ColorField { value = (Color)value },
                { } when type == typeof(Vector3) => new Vector3Field { value = (Vector3)value },
                { } when type == typeof(Vector2) => new Vector2Field { value = (Vector2)value },
                { } when type == typeof(GreyscaleValue) => new GreyscaleField { value = ((GreyscaleValue)value).value },
                { } when type.IsEnum => new EnumField((Enum)value),
                { } when typeof(UnityEngine.Object).IsAssignableFrom(type) => new ObjectField
                {
                    value = (UnityEngine.Object)value,
                    objectType = type
                },
                _ => null
            };
        }

        public void ToggleExposedVariable(Port port, bool isVisible)
        {
            var assetGraphPort = port as AssetGraphPort;
            if (assetGraphPort?.ExposedPropertyContainer != null)
                assetGraphPort.ExposedPropertyContainer.style.display =
                    isVisible ? DisplayStyle.Flex : DisplayStyle.None;
        }

        public override void BuildContextualMenu(ContextualMenuPopulateEvent evt)
        {
            evt.menu.InsertAction(evt.menu.MenuItems().Count, "Open Documentation", OpenDocumentation);
            evt.menu.AppendSeparator();
            base.BuildContextualMenu(evt);
        }

        private void OpenDocumentation(DropdownMenuAction obj)
        {
            if (!string.IsNullOrEmpty(info.DocumentationationURL))
            {
                Application.OpenURL(info.DocumentationationURL);
            }
            else
            {
                Debug.LogWarning($"No documentation URL provided for node {GraphNode.asset.name}!");
            }
        }

        public void SavePosition() => GraphNode.SetPosition(GetPosition());
    }
}