﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Reflection;
using AssetGraph.Core.Attributes;
using AssetGraph.Nodes;
using UnityEditor;
using UnityEditor.Experimental.GraphView;
using UnityEngine;
using UnityEngine.UIElements;

namespace AssetGraph.Core
{
    public struct SearchContextElement
    {
        public object target { get; private set; }
        public string title { get; private set; }

        public Dictionary<int, Type> ImportPortTypes { get; private set; }
        public Dictionary<int, Type> ExportPortTypes { get; private set; }
        public int portID;

        public SearchContextElement(object target, string title, Dictionary<int, Type> importPortTypes,
            Dictionary<int, Type> exportPortTypes, int portID)
        {
            this.target = target;
            this.title = title;
            ImportPortTypes = importPortTypes;
            ExportPortTypes = exportPortTypes;
            this.portID = portID;
        }
    }

    public class AssetGraphSearchProvider : ScriptableObject, ISearchWindowProvider
    {
        public AssetGraphViewWindow graph;
        public VisualElement target;

        public static List<SearchContextElement> elements;
        private Assembly[] assemblies;

        // Create Search Tree for Nodes
        public List<SearchTreeEntry> CreateSearchTree(SearchWindowContext context)
        {
            List<SearchTreeEntry> tree = new List<SearchTreeEntry>();
            tree.Add(new SearchTreeGroupEntry(new GUIContent("Nodes"), 0));

            elements = new List<SearchContextElement>();

            foreach (var type in TypeCache.GetTypesWithAttribute<NodeInfoAttribute>())
            {
                var attr = type.GetCustomAttribute<NodeInfoAttribute>();
                if (attr == null || string.IsNullOrEmpty(attr.MenuItem)) continue;

                var node = Activator.CreateInstance(type);

                var fields = type.GetFields();
                var importPortTypes = GetPortsByAttribute(fields, typeof(Core.Attributes.Input));
                var exportPortTypes = GetPortsByAttribute(fields, typeof(Output));

                elements.Add(new SearchContextElement(node, attr.MenuItem, importPortTypes, exportPortTypes, 0));
            }

            elements.Sort(SortElementsByTitle);

            var groups = new HashSet<string>();

            foreach (var element in elements)
            {
                string[] entryTitle = element.title.Split('/');
                string currentGroup = "";

                for (int i = 0; i < entryTitle.Length - 1; i++)
                {
                    currentGroup += entryTitle[i];
                    if (groups.Add(currentGroup))
                        tree.Add(new SearchTreeGroupEntry(new GUIContent(entryTitle[i]), i + 1));

                    currentGroup += '/';
                }

                tree.Add(new SearchTreeEntry(new GUIContent(entryTitle.Last()))
                {
                    level = entryTitle.Length,
                    userData = element
                });
            }

            return tree;
        }

        // Handle Selection from Search Tree
        public bool OnSelectEntry(SearchTreeEntry SearchTreeEntry, SearchWindowContext context)
        {
            var mousePos =
                graph.ChangeCoordinatesTo(graph, context.screenMousePosition - graph.Window.position.position);
            var graphMousePosition = graph.contentViewContainer.WorldToLocal(mousePos);

            SearchContextElement element = (SearchContextElement)SearchTreeEntry.userData;
            AssetGraphNode node = (AssetGraphNode)element.target;
            node.SetPosition(new Rect(graphMousePosition, new Vector2()));
            graph.Add(node);
            node.asset = graph.asset;

            return true;
        }

        // Create Search Tree for Edge Connections
        public List<SearchTreeEntry> CreateSearchTreeForEdge(SearchWindowContext context, Edge edge)
        {
            var tree = new List<SearchTreeEntry>();
            tree.Add(new SearchTreeGroupEntry(new GUIContent("Compatible Nodes"), 0));

            elements = new List<SearchContextElement>();
            var groups = new HashSet<string>();

            var sourcePort = edge.output ?? edge.input;
            bool isSourceOutput = edge.output != null;
            Type targetType = (sourcePort as AssetGraphPort)?.portType;

            if (targetType == null)
            {
                Debug.LogWarning("Could not determine port type from edge!");
                return tree;
            }

            foreach (var type in TypeCache.GetTypesWithAttribute<NodeInfoAttribute>())
            {
                var attr = type.GetCustomAttribute<NodeInfoAttribute>();
                if (attr == null || string.IsNullOrEmpty(attr.MenuItem)) continue;

                var node = Activator.CreateInstance(type);
                var fields = type.GetFields();

                var inputPorts = GetPortsByAttributeList(fields, typeof(Core.Attributes.Input));
                var outputPorts = GetPortsByAttributeList(fields, typeof(Output));

                var compatiblePorts = isSourceOutput
                    ? inputPorts.Where(port => port.FieldType.IsAssignableFrom(targetType)).ToList()
                    : outputPorts.Where(port => targetType.IsAssignableFrom(port.FieldType)).ToList();

                if (!compatiblePorts.Any()) continue;

                string[] menuPath = attr.MenuItem.Split('/');
                string currentGroupPath = "";

                for (int i = 0; i < menuPath.Length - 1; i++)
                {
                    currentGroupPath += menuPath[i];
                    if (groups.Add(currentGroupPath))
                        tree.Add(new SearchTreeGroupEntry(new GUIContent(menuPath[i]), i + 1));

                    currentGroupPath += "/";
                }

                foreach (var compatiblePort in compatiblePorts)
                {
                    tree.Add(new SearchTreeEntry(new GUIContent($"{attr.Title}: {compatiblePort.Name}"))
                    {
                        level = menuPath.Length,
                        userData = new SearchContextElement(
                            node,
                            attr.Title,
                            inputPorts.ToDictionary(field => inputPorts.IndexOf(field), field => field.FieldType),
                            outputPorts.ToDictionary(field => outputPorts.IndexOf(field), field => field.FieldType),
                            inputPorts.IndexOf(compatiblePort))
                    });
                }
            }

            return tree;
        }

        private static Dictionary<int, Type> GetPortsByAttribute(FieldInfo[] fields, Type attributeType)
        {
            return fields
                .Where(f => f.GetCustomAttribute(attributeType) != null)
                .Select((f, index) => new { f, index })
                .ToDictionary(a => a.index, a => a.f.FieldType);
        }

        private static List<FieldInfo> GetPortsByAttributeList(FieldInfo[] fields, Type attributeType)
        {
            return fields.Where(f => f.GetCustomAttribute(attributeType) != null).ToList();
        }

        private static int SortElementsByTitle(SearchContextElement a, SearchContextElement b)
        {
            return string.Compare(a.title, b.title, StringComparison.Ordinal);
        }

        // CustomSearchProviderForEdge
        public class CustomSearchProviderForEdge : ScriptableObject, ISearchWindowProvider
        {
            private AssetGraphSearchProvider original;
            private Edge edge;
            private AssetGraphPort inputPort;
            private AssetGraphPort outputPort;

            public CustomSearchProviderForEdge(AssetGraphSearchProvider original, Edge edge, AssetGraphPort inputPort,
                AssetGraphPort outputPort)
            {
                Init(original, edge, inputPort, outputPort);
            }

            public void Init(AssetGraphSearchProvider original, Edge edge, AssetGraphPort inputPort,
                AssetGraphPort outputPort)
            {
                this.original = original;
                this.edge = edge;
                this.inputPort = inputPort;
                this.outputPort = outputPort;
            }

            public List<SearchTreeEntry> CreateSearchTree(SearchWindowContext context)
            {
                return original.CreateSearchTreeForEdge(context, edge);
            }

            public bool OnSelectEntry(SearchTreeEntry selectedEntry, SearchWindowContext context)
            {
                var mousePos = original.graph.ChangeCoordinatesTo(original.graph,
                    context.screenMousePosition - original.graph.Window.position.position);
                var graphMousePosition = original.graph.contentViewContainer.WorldToLocal(mousePos);

                var element = (SearchContextElement)selectedEntry.userData;
                var node = (AssetGraphNode)element.target;

                node.SetPosition(new Rect(graphMousePosition, new Vector2()));
                original.graph.Add(node);
                node.asset = original.graph.asset;

                Edge newEdge = new Edge();
                if (inputPort != null)
                    newEdge = inputPort.ConnectTo(original.graph.nodeDictionary[node.ID].OutputPorts[element.portID]);
                else if (outputPort != null)
                    newEdge = outputPort.ConnectTo(original.graph.nodeDictionary[node.ID].InputPorts[element.portID]);

                original.graph.CreateEdge(newEdge);
                original.graph.AddElement(newEdge);

                return true;
            }
        }
    }
}